from __future__ import annotations

import warnings
from collections.abc import Callable

import cv2
import numpy as np
import torch
from torch.utils.data import Dataset

from quadra.utils.imaging import crop_image, keep_aspect_ratio_resize


class ImageClassificationListDataset(Dataset):
    """Standard classification dataset.

    Args:
        samples: List of paths to images to be read
        targets: List of labels, one for every image
            in samples
        class_to_idx: mapping from classes
            to unique indexes.
            Defaults to None.
        resize: Integer specifying the size of
            a first optional resize keeping the aspect ratio: the smaller side
            of the image will be resized to `resize`, while the longer will be
            resized keeping the aspect ratio.
            Defaults to None.
        roi: Optional ROI, with
            (x_upper_left, y_upper_left, x_bottom_right, y_bottom_right).
            Defaults to None.
        transform: Optional Albumentations
            transform.
            Defaults to None.
        rgb: if False, image will be converted in grayscale
        channel: 1 or 3. If rgb is True, then channel will be set at 3.
        allow_missing_label: If set to false warn the user if the dataset contains missing labels
    """

    def __init__(
        self,
        samples: list[str],
        targets: list[str | int],
        class_to_idx: dict | None = None,
        resize: int | None = None,
        roi: tuple[int, int, int, int] | None = None,
        transform: Callable | None = None,
        rgb: bool = True,
        channel: int = 3,
        allow_missing_label: bool | None = False,
    ):
        super().__init__()
        assert len(samples) == len(targets), (
            f"Samples ({len(samples)}) and targets ({len(targets)}) must have the same length"
        )
        # Setting the ROI
        self.roi = roi

        # Keep-Aspect-Ratio resize
        self.resize = resize

        if not allow_missing_label and None in targets:
            warnings.warn(
                (
                    "Dataset contains empty targets but allow_missing_label is set to False, "
                    "be careful because None labels will not work inside Dataloaders"
                ),
                UserWarning,
                stacklevel=2,
            )

        targets = [-1 if target is None else target for target in targets]
        # Data
        self.x = np.array(samples)
        self.y = np.array(targets)

        if class_to_idx is None:
            unique_targets = np.unique(targets)
            class_to_idx = {c: i for i, c in enumerate(unique_targets)}

        self.class_to_idx = class_to_idx
        self.idx_to_class = {v: k for k, v in class_to_idx.items()}
        self.samples = [
            (path, self.class_to_idx[self.y[i]] if (self.y[i] != -1 and self.y[i] != "-1") else -1)
            for i, path in enumerate(self.x)
        ]

        self.rgb = rgb
        self.channel = 3 if rgb else channel

        self.transform = transform

    def __getitem__(self, idx) -> tuple[np.ndarray, np.ndarray]:
        path, y = self.samples[idx]

        # Load image
        x = cv2.imread(str(path))
        if self.rgb:
            x = cv2.cvtColor(x, cv2.COLOR_BGR2RGB)
        else:
            x = cv2.cvtColor(x, cv2.COLOR_BGR2GRAY)
            x = cv2.cvtColor(x, cv2.COLOR_GRAY2RGB)

        if self.channel == 1:
            x = x[:, :, 0]

        # Crop with ROI
        if self.roi:
            x = crop_image(x, self.roi)

        # Resize keeping aspect ratio
        if self.resize:
            x = keep_aspect_ratio_resize(x, self.resize)

        if self.transform:
            aug = self.transform(image=x)
            x = aug["image"]

        return x, y

    def __len__(self):
        return len(self.samples)


class ClassificationDataset(ImageClassificationListDataset):
    """Custom Classification Dataset.

    Args:
        samples: List of paths to images
        targets: List of targets
        class_to_idx: Defaults to None.
        resize: Resize image to this size. Defaults to None.
        roi: Region of interest. Defaults to None.
        transform: transform function. Defaults to None.
        rgb: Use RGB space
        channel: Number of channels. Defaults to 3.
        random_padding: Random padding. Defaults to False.
        circular_crop: Circular crop. Defaults to False.
    """

    def __init__(
        self,
        samples: list[str],
        targets: list[str | int],
        class_to_idx: dict | None = None,
        resize: int | None = None,
        roi: tuple[int, int, int, int] | None = None,
        transform: Callable | None = None,
        rgb: bool = True,
        channel: int = 3,
        random_padding: bool = False,
        circular_crop: bool = False,
    ):
        super().__init__(samples, targets, class_to_idx, resize, roi, transform, rgb, channel)
        if transform is None:
            self.transform = None

        self.random_padding = random_padding
        self.circular_crop = circular_crop

    def __getitem__(self, idx):
        path, y = self.samples[idx]
        path = str(path)

        # Load image
        x = cv2.imread(path)
        if self.rgb:
            x = cv2.cvtColor(x, cv2.COLOR_BGR2RGB)
        else:
            x = cv2.cvtColor(x, cv2.COLOR_BGR2GRAY)
            x = cv2.cvtColor(x, cv2.COLOR_GRAY2RGB)

        if self.transform is not None:
            aug = self.transform(image=x)
            x = aug["image"]

        if self.channel == 1:
            x = x[:1]

        return x, y


class MultilabelClassificationDataset(torch.utils.data.Dataset):
    """Custom MultilabelClassification Dataset.

    Args:
        samples: list of paths to images.
        targets: array of multiple targets per sample. The array must be a one-hot enoding.
            It must have a shape of (n_samples, n_targets).
        class_to_idx: Defaults to None.
        transform: transform function. Defaults to None.
        rgb: Use RGB space
    """

    def __init__(
        self,
        samples: list[str],
        targets: np.ndarray,
        class_to_idx: dict | None = None,
        transform: Callable | None = None,
        rgb: bool = True,
    ):
        super().__init__()
        assert len(samples) == len(targets), (
            f"Samples ({len(samples)}) and targets ({len(targets)}) must have the same length"
        )

        # Data
        self.x = samples
        self.y = targets

        # Class to idx and the other way around
        if class_to_idx is None:
            unique_targets = targets.shape[1]
            class_to_idx = {c: i for i, c in enumerate(range(unique_targets))}
        self.class_to_idx = class_to_idx
        self.idx_to_class = {v: k for k, v in class_to_idx.items()}
        self.samples = list(zip(self.x, self.y, strict=False))
        self.rgb = rgb
        self.transform = transform

    def __getitem__(self, idx):
        path, y = self.samples[idx]
        path = str(path)

        # Load image
        x = cv2.imread(path)
        if self.rgb:
            x = cv2.cvtColor(x, cv2.COLOR_BGR2RGB)
        else:
            x = cv2.cvtColor(x, cv2.COLOR_BGR2GRAY)
            x = cv2.cvtColor(x, cv2.COLOR_GRAY2RGB)

        if self.transform is not None:
            aug = self.transform(image=x)
            x = aug["image"]

        return x, torch.from_numpy(y).float()

    def __len__(self):
        return len(self.samples)



from __future__ import annotations

import json
import os
from pathlib import Path
from typing import Any, cast

import hydra
import torch
from joblib import dump, load
from omegaconf import DictConfig, OmegaConf
from sklearn.base import ClassifierMixin

from quadra.datamodules import PatchSklearnClassificationDataModule
from quadra.datasets.patch import PatchSklearnClassificationTrainDataset
from quadra.models.base import ModelSignatureWrapper
from quadra.models.evaluation import BaseEvaluationModel
from quadra.tasks.base import Evaluation, Task
from quadra.trainers.classification import SklearnClassificationTrainer
from quadra.utils import utils
from quadra.utils.classification import automatic_batch_size_computation
from quadra.utils.evaluation import automatic_datamodule_batch_size
from quadra.utils.export import export_model, import_deployment_model
from quadra.utils.patch import RleEncoder, compute_patch_metrics, save_classification_result
from quadra.utils.patch.dataset import PatchDatasetFileFormat

log = utils.get_logger(__name__)


class PatchSklearnClassification(Task[PatchSklearnClassificationDataModule]):
    """Patch classification using torch backbone for feature extraction and sklearn to learn a linear classifier.

    Args:
        config: The experiment configuration
        device: The device to use
        output: Dictionary defining which kind of outputs to generate. Defaults to None.
        automatic_batch_size: Whether to automatically find the largest batch size that fits in memory.
        half_precision: Whether to use half precision.
    """

    def __init__(
        self,
        config: DictConfig,
        output: DictConfig,
        device: str,
        automatic_batch_size: DictConfig,
        half_precision: bool = False,
    ):
        super().__init__(config=config)
        self.device: str = device
        self.output: DictConfig = output
        self.return_polygon: bool = True
        self.reconstruction_results: dict[str, Any]
        self._backbone: ModelSignatureWrapper
        self._trainer: SklearnClassificationTrainer
        self._model: ClassifierMixin
        self.metadata: dict[str, Any] = {
            "test_confusion_matrix": [],
            "test_accuracy": [],
            "test_results": [],
            "test_labels": [],
        }
        self.export_folder: str = "deployment_model"
        self.automatic_batch_size = automatic_batch_size
        self.half_precision = half_precision

    @property
    def model(self) -> ClassifierMixin:
        """sklearn.base.ClassifierMixin: The model."""
        return self._model

    @model.setter
    def model(self, model_config: DictConfig):
        """sklearn.base.ClassifierMixin: The model."""
        log.info("Instantiating model <%s>", model_config["_target_"])
        self._model = hydra.utils.instantiate(model_config)

    @property
    def backbone(self) -> ModelSignatureWrapper:
        """Backbone: The backbone."""
        return self._backbone

    @backbone.setter
    def backbone(self, backbone_config):
        """Load backbone."""
        if backbone_config.metadata.get("checkpoint"):
            log.info("Loading backbone from <%s>", backbone_config.metadata.checkpoint)
            self._backbone = torch.load(backbone_config.metadata.checkpoint)
        else:
            log.info("Loading backbone from <%s>", backbone_config.model["_target_"])
            self._backbone = hydra.utils.instantiate(backbone_config.model)

        self._backbone = ModelSignatureWrapper(self._backbone)
        self._backbone.eval()
        if self.half_precision:
            if self.device == "cpu":
                raise ValueError("Half precision is not supported on CPU")
            self._backbone.half()
        self._backbone = self._backbone.to(self.device)

    def prepare(self) -> None:
        """Prepare the experiment."""
        self.datamodule = self.config.datamodule
        self.backbone = self.config.backbone
        self.model = self.config.model

        if not self.automatic_batch_size.disable and self.device != "cpu":
            self.datamodule.batch_size = automatic_batch_size_computation(
                datamodule=self.datamodule,
                backbone=self.backbone,
                starting_batch_size=self.automatic_batch_size.starting_batch_size,
            )

        self.trainer = self.config.trainer

    @property
    def trainer(self) -> SklearnClassificationTrainer:
        """Trainer: The trainer."""
        return self._trainer

    @trainer.setter
    def trainer(self, trainer_config: DictConfig) -> None:
        """Trainer: The trainer."""
        log.info("Instantiating trainer <%s>", trainer_config["_target_"])
        trainer = hydra.utils.instantiate(trainer_config, backbone=self.backbone, classifier=self.model)
        self._trainer = trainer

    def train(self) -> None:
        """Train the model."""
        log.info("Starting training...!")
        # prepare_data() must be explicitly called if the task does not include a lightining training
        self.datamodule.prepare_data()
        self.datamodule.setup(stage="fit")
        class_to_keep = None
        if hasattr(self.datamodule, "class_to_skip_training") and self.datamodule.class_to_skip_training is not None:
            class_to_keep = [x for x in self.datamodule.class_to_idx if x not in self.datamodule.class_to_skip_training]

        self.model = self.config.model
        self.trainer.change_classifier(self.model)
        train_dataloader = self.datamodule.train_dataloader()
        val_dataloader = self.datamodule.val_dataloader()
        train_dataset = cast(PatchSklearnClassificationTrainDataset, train_dataloader.dataset)
        self.trainer.fit(train_dataloader=train_dataloader)
        _, pd_cm, accuracy, res, _ = self.trainer.test(
            test_dataloader=val_dataloader,
            class_to_keep=class_to_keep,
            idx_to_class=train_dataset.idx_to_class,
            predict_proba=True,
        )

        # save results
        self.metadata["test_confusion_matrix"] = pd_cm
        self.metadata["test_accuracy"] = accuracy
        self.metadata["test_results"] = res
        self.metadata["test_labels"] = [
            train_dataset.idx_to_class[i] if i != -1 else "N/A" for i in res["real_label"].unique().tolist()
        ]

    def generate_report(self) -> None:
        """Generate the report for the task."""
        log.info("Generating report!")
        os.makedirs(self.output.folder, exist_ok=True)

        c_matrix = self.metadata["test_confusion_matrix"]
        idx_to_class = {v: k for k, v in self.datamodule.class_to_idx.items()}

        datamodule: PatchSklearnClassificationDataModule = self.datamodule
        val_img_info: list[PatchDatasetFileFormat] = datamodule.info.val_files
        for img_info in val_img_info:
            if not os.path.isabs(img_info.image_path):
                img_info.image_path = os.path.join(datamodule.data_path, img_info.image_path)
            if img_info.mask_path is not None and not os.path.isabs(img_info.mask_path):
                img_info.mask_path = os.path.join(datamodule.data_path, img_info.mask_path)

        false_region_bad, false_region_good, true_region_bad, reconstructions = compute_patch_metrics(
            test_img_info=val_img_info,
            test_results=self.metadata["test_results"],
            patch_num_h=datamodule.info.patch_number[0] if datamodule.info.patch_number is not None else None,
            patch_num_w=datamodule.info.patch_number[1] if datamodule.info.patch_number is not None else None,
            patch_h=datamodule.info.patch_size[0] if datamodule.info.patch_size is not None else None,
            patch_w=datamodule.info.patch_size[1] if datamodule.info.patch_size is not None else None,
            overlap=datamodule.info.overlap,
            idx_to_class=idx_to_class,
            return_polygon=self.return_polygon,
            patch_reconstruction_method=self.output.reconstruction_method,
            annotated_good=datamodule.info.annotated_good,
        )

        self.reconstruction_results = {
            "false_region_bad": false_region_bad,
            "false_region_good": false_region_good,
            "true_region_bad": true_region_bad,
            "reconstructions": reconstructions,
            "reconstructions_type": "polygon" if self.return_polygon else "rle",
            "patch_reconstruction_method": self.output.reconstruction_method,
        }

        with open("reconstruction_results.json", "w") as f:
            json.dump(
                self.reconstruction_results,
                f,
                cls=RleEncoder,
            )

        if hasattr(self.datamodule, "class_to_skip_training") and self.datamodule.class_to_skip_training is not None:
            ignore_classes = [self.datamodule.class_to_idx[x] for x in self.datamodule.class_to_skip_training]
        else:
            ignore_classes = None
        val_dataloader = self.datamodule.val_dataloader()
        save_classification_result(
            results=self.metadata["test_results"],
            output_folder=self.output.folder,
            confusion_matrix=c_matrix,
            accuracy=self.metadata["test_accuracy"],
            test_dataloader=val_dataloader,
            config=self.config,
            output=self.output,
            reconstructions=reconstructions,
            ignore_classes=ignore_classes,
        )

    def export(self) -> None:
        """Generate deployment model for the task."""
        input_shapes = self.config.export.input_shapes

        idx_to_class = {v: k for k, v in self.datamodule.class_to_idx.items()}

        model_json, export_paths = export_model(
            config=self.config,
            model=self.backbone,
            export_folder=self.export_folder,
            half_precision=self.half_precision,
            input_shapes=input_shapes,
            idx_to_class=idx_to_class,
            pytorch_model_type="backbone",
        )

        if len(export_paths) > 0:
            dataset_info = self.datamodule.info

            horizontal_patches = dataset_info.patch_number[1] if dataset_info.patch_number is not None else None
            vertical_patches = dataset_info.patch_number[0] if dataset_info.patch_number is not None else None
            patch_height = dataset_info.patch_size[0] if dataset_info.patch_size is not None else None
            patch_width = dataset_info.patch_size[1] if dataset_info.patch_size is not None else None
            overlap = dataset_info.overlap

            model_json.update(
                {
                    "horizontal_patches": horizontal_patches,
                    "vertical_patches": vertical_patches,
                    "patch_height": patch_height,
                    "patch_width": patch_width,
                    "overlap": overlap,
                    "reconstruction_method": self.output.reconstruction_method,
                    "class_to_skip": self.datamodule.class_to_skip_training,
                }
            )

            with open(os.path.join(self.export_folder, "model.json"), "w") as f:
                json.dump(model_json, f, cls=utils.HydraEncoder)

        dump(self.model, os.path.join(self.export_folder, "classifier.joblib"))

    def execute(self) -> None:
        """Execute the experiment and all the steps."""
        self.prepare()
        self.train()
        if self.output.report:
            self.generate_report()
        if self.config.export is not None and len(self.config.export.types) > 0:
            self.export()
        self.finalize()


class PatchSklearnTestClassification(Evaluation[PatchSklearnClassificationDataModule]):
    """Perform a test of an already trained classification model.

    Args:
        config: The experiment configuration
        output: where to save resultss
        model_path: path to trained model from PatchSklearnClassification task.
        device: the device where to run the model (cuda or cpu). Defaults to 'cpu'.
    """

    def __init__(
        self,
        config: DictConfig,
        output: DictConfig,
        model_path: str,
        device: str = "cpu",
    ):
        super().__init__(config=config, model_path=model_path, device=device)
        self.output = output
        self._backbone: BaseEvaluationModel
        self._classifier: ClassifierMixin
        self.class_to_idx: dict[str, int]
        self.idx_to_class: dict[int, str]
        self.metadata: dict[str, Any] = {
            "test_confusion_matrix": None,
            "test_accuracy": None,
            "test_results": None,
            "test_labels": None,
        }
        self.class_to_skip: list[str] | None = []
        self.reconstruction_results: dict[str, Any]
        self.return_polygon: bool = True

    def prepare(self) -> None:
        """Prepare the experiment."""
        super().prepare()

        idx_to_class = {}
        class_to_idx = {}
        for k, v in self.model_data["classes"].items():
            idx_to_class[int(k)] = v
            class_to_idx[v] = int(k)

        self.idx_to_class = idx_to_class
        self.class_to_idx = class_to_idx
        self.config.datamodule.class_to_idx = class_to_idx

        self.datamodule = self.config.datamodule
        # Configure trainer
        self.trainer = self.config.trainer

        # prepare_data() must be explicitly called because there is no lightning training
        self.datamodule.prepare_data()
        self.datamodule.setup(stage="test")

    @automatic_datamodule_batch_size(batch_size_attribute_name="batch_size")
    def test(self) -> None:
        """Run the test."""
        test_dataloader = self.datamodule.test_dataloader()

        self.class_to_skip = self.model_data["class_to_skip"] if hasattr(self.model_data, "class_to_skip") else None
        class_to_keep = None

        if self.class_to_skip is not None:
            class_to_keep = [x for x in self.datamodule.class_to_idx if x not in self.class_to_skip]
        _, pd_cm, accuracy, res, _ = self.trainer.test(
            test_dataloader=test_dataloader,
            idx_to_class=self.idx_to_class,
            predict_proba=True,
            class_to_keep=class_to_keep,
        )

        # save results
        self.metadata["test_confusion_matrix"] = pd_cm
        self.metadata["test_accuracy"] = accuracy
        self.metadata["test_results"] = res
        self.metadata["test_labels"] = [
            self.idx_to_class[i] if i != -1 else "N/A" for i in res["real_label"].unique().tolist()
        ]

    @property
    def deployment_model(self):
        """Deployment model."""
        return None

    @deployment_model.setter
    def deployment_model(self, model_path: str):
        """Set backbone and classifier."""
        self.backbone = model_path  # type: ignore[assignment]
        # Load classifier
        self.classifier = os.path.join(Path(model_path).parent, "classifier.joblib")

    @property
    def classifier(self) -> ClassifierMixin:
        """Classifier: The classifier."""
        return self._classifier

    @classifier.setter
    def classifier(self, classifier_path: str) -> None:
        """Load classifier."""
        self._classifier = load(classifier_path)

    @property
    def backbone(self) -> BaseEvaluationModel:
        """Backbone: The backbone."""
        return self._backbone

    @backbone.setter
    def backbone(self, model_path: str) -> None:
        """Load backbone."""
        file_extension = os.path.splitext(model_path)[1]

        model_architecture = None
        if file_extension == ".pth":
            backbone_config_path = os.path.join(Path(model_path).parent, "model_config.yaml")
            log.info("Loading backbone from config")
            backbone_config = OmegaConf.load(backbone_config_path)

            if backbone_config.metadata.get("checkpoint"):
                log.info("Loading backbone from <%s>", backbone_config.metadata.checkpoint)
                model_architecture = torch.load(backbone_config.metadata.checkpoint)
            else:
                log.info("Loading backbone from <%s>", backbone_config.model["_target_"])
                model_architecture = hydra.utils.instantiate(backbone_config.model)

        self._backbone = import_deployment_model(
            model_path=model_path,
            device=self.device,
            inference_config=self.config.inference,
            model_architecture=model_architecture,
        )

    @property
    def trainer(self) -> SklearnClassificationTrainer:
        """Trainer: The trainer."""
        return self._trainer

    @trainer.setter
    def trainer(self, trainer_config: DictConfig) -> None:
        """Trainer: The trainer."""
        log.info("Instantiating trainer <%s>", trainer_config["_target_"])

        if self.backbone.training:
            self.backbone.eval()

        trainer = hydra.utils.instantiate(trainer_config, backbone=self.backbone, classifier=self.classifier)
        self._trainer = trainer

    def generate_report(self) -> None:
        """Generate a report for the task."""
        log.info("Generating report!")
        os.makedirs(self.output.folder, exist_ok=True)

        c_matrix = self.metadata["test_confusion_matrix"]
        idx_to_class = {v: k for k, v in self.datamodule.class_to_idx.items()}

        datamodule: PatchSklearnClassificationDataModule = self.datamodule
        test_img_info = datamodule.info.test_files
        for img_info in test_img_info:
            if not os.path.isabs(img_info.image_path):
                img_info.image_path = os.path.join(datamodule.data_path, img_info.image_path)
            if img_info.mask_path is not None and not os.path.isabs(img_info.mask_path):
                img_info.mask_path = os.path.join(datamodule.data_path, img_info.mask_path)

        false_region_bad, false_region_good, true_region_bad, reconstructions = compute_patch_metrics(
            test_img_info=test_img_info,
            test_results=self.metadata["test_results"],
            patch_num_h=datamodule.info.patch_number[0] if datamodule.info.patch_number is not None else None,
            patch_num_w=datamodule.info.patch_number[1] if datamodule.info.patch_number is not None else None,
            patch_h=datamodule.info.patch_size[0] if datamodule.info.patch_size is not None else None,
            patch_w=datamodule.info.patch_size[1] if datamodule.info.patch_size is not None else None,
            overlap=datamodule.info.overlap,
            idx_to_class=idx_to_class,
            return_polygon=self.return_polygon,
            patch_reconstruction_method=self.output.reconstruction_method,
            annotated_good=datamodule.info.annotated_good,
        )

        self.reconstruction_results = {
            "false_region_bad": false_region_bad,
            "false_region_good": false_region_good,
            "true_region_bad": true_region_bad,
            "reconstructions": reconstructions,
            "reconstructions_type": "polygon" if self.return_polygon else "rle",
            "patch_reconstruction_method": self.output.reconstruction_method,
        }

        with open("reconstruction_results.json", "w") as f:
            json.dump(
                self.reconstruction_results,
                f,
                cls=RleEncoder,
            )

        if self.class_to_skip is not None:
            ignore_classes = [datamodule.class_to_idx[x] for x in self.class_to_skip]
        else:
            ignore_classes = None
        test_dataloader = self.datamodule.test_dataloader()
        save_classification_result(
            results=self.metadata["test_results"],
            output_folder=self.output.folder,
            confusion_matrix=c_matrix,
            accuracy=self.metadata["test_accuracy"],
            test_dataloader=test_dataloader,
            config=self.config,
            output=self.output,
            reconstructions=reconstructions,
            ignore_classes=ignore_classes,
        )

    def execute(self) -> None:
        """Execute the experiment and all the steps."""
        self.prepare()
        self.test()
        if self.output.report:
            self.generate_report()
        self.finalize()



import torch
import torch.nn.functional as F
from torch import nn


def byol_regression_loss(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """Byol regression loss
    Args:
        x: tensor
        y: tensor.

    Returns:
        tensor
    """
    x = F.normalize(x, dim=-1)
    y = F.normalize(y, dim=-1)
    return 2 - 2 * (x * y).sum(dim=1).mean()


class BYOLRegressionLoss(nn.Module):
    """BYOL regression loss module."""

    def forward(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
    ) -> torch.Tensor:
        """Compute the BYOL regression loss.

        Args:
            x: First Tensor
            y: Second Tensor

        Returns:
            BYOL regression loss
        """
        return byol_regression_loss(x, y)



from __future__ import annotations

import contextlib
import os
from collections.abc import Sequence
from typing import Any, Literal, TypeVar, cast

import torch
from anomalib.models.cflow import CflowLightning
from omegaconf import DictConfig, ListConfig, OmegaConf
from torch import nn

from quadra.models.base import ModelSignatureWrapper
from quadra.models.evaluation import (
    BaseEvaluationModel,
    ONNXEvaluationModel,
    TorchEvaluationModel,
    TorchscriptEvaluationModel,
)
from quadra.utils.logger import get_logger

try:
    import onnx  # noqa
    from onnxsim import simplify as onnx_simplify  # noqa
    from onnxconverter_common import auto_convert_mixed_precision  # noqa

    ONNX_AVAILABLE = True
except ImportError:
    ONNX_AVAILABLE = False

log = get_logger(__name__)

BaseDeploymentModelT = TypeVar("BaseDeploymentModelT", bound=BaseEvaluationModel)


def generate_torch_inputs(
    input_shapes: list[Any],
    device: str | torch.device,
    half_precision: bool = False,
    dtype: torch.dtype = torch.float32,
    batch_size: int = 1,
) -> list[Any] | tuple[Any, ...] | torch.Tensor:
    """Given a list of input shapes that can contain either lists, tuples or dicts, with tuples being the input shapes
    of the model, generate a list of torch tensors with the given device and dtype.
    """
    inp = None

    if isinstance(input_shapes, ListConfig | DictConfig):
        input_shapes = OmegaConf.to_container(input_shapes)

    if isinstance(input_shapes, list):
        if any(isinstance(inp, Sequence | dict) for inp in input_shapes):
            return [generate_torch_inputs(inp, device, half_precision, dtype) for inp in input_shapes]

        # Base case
        inp = torch.randn((batch_size, *input_shapes), dtype=dtype, device=device)

    if isinstance(input_shapes, dict):
        return {k: generate_torch_inputs(v, device, half_precision, dtype) for k, v in input_shapes.items()}

    if isinstance(input_shapes, tuple):
        if any(isinstance(inp, Sequence | dict) for inp in input_shapes):
            # The tuple contains a list, tuple or dict
            return tuple(generate_torch_inputs(inp, device, half_precision, dtype) for inp in input_shapes)

        # Base case
        inp = torch.randn((batch_size, *input_shapes), dtype=dtype, device=device)

    if inp is None:
        raise RuntimeError("Something went wrong during model export, unable to parse input shapes")

    if half_precision:
        inp = inp.half()

    return inp


def extract_torch_model_inputs(
    model: nn.Module | ModelSignatureWrapper,
    input_shapes: list[Any] | None = None,
    half_precision: bool = False,
    batch_size: int = 1,
) -> tuple[list[Any] | tuple[Any, ...] | torch.Tensor, list[Any]] | None:
    """Extract the input shapes for the given model and generate a list of torch tensors with the
    given device and dtype.

    Args:
        model: Module or ModelSignatureWrapper
        input_shapes: Inputs shapes
        half_precision: If True, the model will be exported with half precision
        batch_size: Batch size for the input shapes
    """
    if isinstance(model, ModelSignatureWrapper) and input_shapes is None:
        input_shapes = model.input_shapes

    if input_shapes is None:
        log.warning(
            "Input shape is None, can not trace model! Please provide input_shapes in the task export configuration."
        )
        return None

    if half_precision:
        # TODO: This doesn't support bfloat16!!
        inp = generate_torch_inputs(
            input_shapes=input_shapes, device="cuda:0", half_precision=True, dtype=torch.float16, batch_size=batch_size
        )
    else:
        inp = generate_torch_inputs(
            input_shapes=input_shapes, device="cpu", half_precision=False, dtype=torch.float32, batch_size=batch_size
        )

    return inp, input_shapes


@torch.inference_mode()
def export_torchscript_model(
    model: nn.Module,
    output_path: str,
    input_shapes: list[Any] | None = None,
    half_precision: bool = False,
    model_name: str = "model.pt",
    example_inputs: list[torch.Tensor] | tuple[torch.Tensor, ...] | torch.Tensor | None = None,
) -> tuple[str, Any] | None:
    """Export a PyTorch model with TorchScript.

    Args:
        model: PyTorch model to be exported
        input_shapes: Inputs shape for tracing
        output_path: Path to save the model
        half_precision: If True, the model will be exported with half precision
        model_name: Name of the exported model
        example_inputs: If provided use this to evaluate the model instead of generating random inputs, it's expected to
            be a list of tensors or a single tensor without batch dimension

    Returns:
        If the model is exported successfully, the path to the model and the input shape are returned.

    """
    if isinstance(model, CflowLightning):
        log.warning("Exporting cflow model with torchscript is not supported yet.")
        return None

    model.eval()
    if half_precision:
        model.to("cuda:0")
        model = model.half()
    else:
        model.cpu()

    batch_size = 1
    model_inputs: tuple[list[Any] | tuple[Any, ...] | torch.Tensor, list[Any]] | None
    if example_inputs is not None:
        if isinstance(example_inputs, Sequence):
            model_input_tensors = []
            model_input_shapes = []

            for example_input in example_inputs:
                new_inp = example_input.to(
                    device="cuda:0" if half_precision else "cpu",
                    dtype=torch.float16 if half_precision else torch.float32,
                )
                new_inp = new_inp.unsqueeze(0).repeat(batch_size, *(1 for x in new_inp.shape))
                model_input_tensors.append(new_inp)
                model_input_shapes.append(new_inp[0].shape)

            model_inputs = (model_input_tensors, [model_input_shapes])
        else:
            new_inp = example_inputs.to(
                device="cuda:0" if half_precision else "cpu",
                dtype=torch.float16 if half_precision else torch.float32,
            )
            new_inp = new_inp.unsqueeze(0).repeat(batch_size, *(1 for x in new_inp.shape))
            model_inputs = (new_inp, [new_inp[0].shape])
    else:
        model_inputs = extract_torch_model_inputs(model, input_shapes, half_precision)

    if model_inputs is None:
        return None

    if isinstance(model, ModelSignatureWrapper):
        model = model.instance

    inp, input_shapes = model_inputs

    try:
        try:
            model_jit = torch.jit.trace(model, inp)
        except RuntimeError as e:
            log.warning("Standard tracing failed with exception %s, attempting tracing with strict=False", e)
            model_jit = torch.jit.trace(model, inp, strict=False)

        os.makedirs(output_path, exist_ok=True)

        model_path = os.path.join(output_path, model_name)
        model_jit.save(model_path)

        log.info("Torchscript model saved to %s", os.path.join(os.getcwd(), model_path))

        return os.path.join(os.getcwd(), model_path), input_shapes
    except Exception as e:
        log.debug("Failed to export torchscript model with exception: %s", e)
        return None


@torch.inference_mode()
def export_onnx_model(
    model: nn.Module,
    output_path: str,
    onnx_config: DictConfig,
    input_shapes: list[Any] | None = None,
    half_precision: bool = False,
    model_name: str = "model.onnx",
    example_inputs: list[torch.Tensor] | tuple[torch.Tensor, ...] | torch.Tensor | None = None,
    rtol: float = 0.01,
    atol: float = 5e-3,
) -> tuple[str, Any] | None:
    """Export a PyTorch model with ONNX.

    Args:
        model: PyTorch model to be exported
        output_path: Path to save the model
        input_shapes: Input shapes for tracing
        onnx_config: ONNX export configuration
        half_precision: If True, the model will be exported with half precision
        model_name: Name of the exported model
        example_inputs: If provided use this to evaluate the model instead of generating random inputs, it's expected to
            be a list of tensors or a single tensor without batch dimension
        rtol: Relative tolerance for the ONNX safe export in fp16
        atol: Absolute tolerance for the ONNX safe export in fp16
    """
    if not ONNX_AVAILABLE:
        log.warning("ONNX is not installed, can not export model in this format.")
        log.warning("Please install ONNX capabilities for quadra with: poetry install -E onnx")
        return None

    model.eval()
    if half_precision:
        model.to("cuda:0")
        model = model.half()
    else:
        model.cpu()

    if hasattr(onnx_config, "fixed_batch_size") and onnx_config.fixed_batch_size is not None:
        batch_size = onnx_config.fixed_batch_size
    else:
        batch_size = 1

    model_inputs: tuple[list[Any] | tuple[Any, ...] | torch.Tensor, list[Any]] | None
    if example_inputs is not None:
        if isinstance(example_inputs, Sequence):
            model_input_tensors = []
            model_input_shapes = []

            for example_input in example_inputs:
                new_inp = example_input.to(
                    device="cuda:0" if half_precision else "cpu",
                    dtype=torch.float16 if half_precision else torch.float32,
                )
                new_inp = new_inp.unsqueeze(0).repeat(batch_size, *(1 for x in new_inp.shape))
                model_input_tensors.append(new_inp)
                model_input_shapes.append(new_inp[0].shape)

            model_inputs = (model_input_tensors, [model_input_shapes])
        else:
            new_inp = example_inputs.to(
                device="cuda:0" if half_precision else "cpu",
                dtype=torch.float16 if half_precision else torch.float32,
            )
            new_inp = new_inp.unsqueeze(0).repeat(batch_size, *(1 for x in new_inp.shape))
            model_inputs = ([new_inp], [new_inp[0].shape])
    else:
        model_inputs = extract_torch_model_inputs(model, input_shapes, half_precision)

    if model_inputs is None:
        return None

    if isinstance(model, ModelSignatureWrapper):
        model = model.instance

    inp, input_shapes = model_inputs

    os.makedirs(output_path, exist_ok=True)

    model_path = os.path.join(output_path, model_name)

    input_names = onnx_config.input_names if hasattr(onnx_config, "input_names") else None

    if input_names is None:
        input_names = []
        for i, _ in enumerate(inp):
            input_names.append(f"input_{i}")

    output = [model(*inp)]
    output_names = onnx_config.output_names if hasattr(onnx_config, "output_names") else None

    if output_names is None:
        output_names = []
        for i, _ in enumerate(output):
            output_names.append(f"output_{i}")

    dynamic_axes = onnx_config.dynamic_axes if hasattr(onnx_config, "dynamic_axes") else None

    if hasattr(onnx_config, "fixed_batch_size") and onnx_config.fixed_batch_size is not None:
        dynamic_axes = None
    elif dynamic_axes is None:
        dynamic_axes = {}
        for i, _ in enumerate(input_names):
            dynamic_axes[input_names[i]] = {0: "batch_size"}

        for i, _ in enumerate(output_names):
            dynamic_axes[output_names[i]] = {0: "batch_size"}

    modified_onnx_config = cast(dict[str, Any], OmegaConf.to_container(onnx_config, resolve=True))

    modified_onnx_config["input_names"] = input_names
    modified_onnx_config["output_names"] = output_names
    modified_onnx_config["dynamic_axes"] = dynamic_axes

    simplify = modified_onnx_config.pop("simplify", False)
    _ = modified_onnx_config.pop("fixed_batch_size", None)

    if len(inp) == 1:
        inp = inp[0]

    if isinstance(inp, list):
        inp = tuple(inp)  # onnx doesn't like lists representing tuples of inputs
    elif isinstance(inp, torch.Tensor):
        inp = (inp,)

    if isinstance(inp, dict):
        raise ValueError("ONNX export does not support model with dict inputs")

    try:
        torch.onnx.export(model=model, args=inp, f=model_path, **modified_onnx_config)

        onnx_model = onnx.load(model_path)
        # Check if ONNX model is valid
        onnx.checker.check_model(onnx_model)
    except Exception as e:
        log.debug("ONNX export failed with error: %s", e)
        return None

    log.info("ONNX model saved to %s", os.path.join(os.getcwd(), model_path))

    if half_precision:
        is_export_ok = _safe_export_half_precision_onnx(
            model=model,
            export_model_path=model_path,
            inp=inp,
            onnx_config=onnx_config,
            input_shapes=input_shapes,
            input_names=input_names,
            rtol=rtol,
            atol=atol,
        )

        if not is_export_ok:
            return None

    if simplify:
        log.info("Attempting to simplify ONNX model")
        onnx_model = onnx.load(model_path)

        try:
            simplified_model, check = onnx_simplify(onnx_model)
        except Exception as e:
            log.debug("ONNX simplification failed with error: %s", e)
            check = False

        if not check:
            log.warning("Something failed during model simplification, only original ONNX model will be exported")
        else:
            model_filename, model_extension = os.path.splitext(model_name)
            model_name = f"{model_filename}_simplified{model_extension}"
            model_path = os.path.join(output_path, model_name)
            onnx.save(simplified_model, model_path)
            log.info("Simplified ONNX model saved to %s", os.path.join(os.getcwd(), model_path))

    return os.path.join(os.getcwd(), model_path), input_shapes


def _safe_export_half_precision_onnx(
    model: nn.Module,
    export_model_path: str,
    inp: torch.Tensor | list[torch.Tensor] | tuple[torch.Tensor, ...],
    onnx_config: DictConfig,
    input_shapes: list[Any],
    input_names: list[str],
    rtol: float = 0.01,
    atol: float = 5e-3,
) -> bool:
    """Check that the exported half precision ONNX model does not contain NaN values. If it does, attempt to export
    the model with a more stable export and overwrite the original model.

    Args:
        model: PyTorch model to be exported
        export_model_path: Path to save the model
        inp: Input tensors for the model
        onnx_config: ONNX export configuration
        input_shapes: Input shapes for the model
        input_names: Input names for the model
        rtol: Relative tolerance to evaluate the model
        atol: Absolute tolerance to evaluate the model

    Returns:
        True if the model is stable or it was possible to export a more stable model, False otherwise.
    """
    test_fp_16_model: BaseEvaluationModel = import_deployment_model(
        export_model_path, OmegaConf.create({"onnx": {}}), "cuda:0"
    )
    if not isinstance(inp, Sequence):
        inp = [inp]

    test_output = test_fp_16_model(*inp)

    if not isinstance(test_output, Sequence):
        test_output = [test_output]

    # Check if there are nan values in any of the outputs
    is_broken_model = any(torch.isnan(out).any() for out in test_output)

    if is_broken_model:
        try:
            log.warning(
                "The exported half precision ONNX model contains NaN values, attempting with a more stable export..."
            )
            # Cast back the fp16 model to fp32 to simulate the export with fp32
            model = model.float()
            log.info("Starting to export model in full precision")
            export_output = export_onnx_model(
                model=model,
                output_path=os.path.dirname(export_model_path),
                # Force to not simplify fp32 model
                onnx_config=DictConfig({**onnx_config, "simplify": False}),
                input_shapes=input_shapes,
                half_precision=False,
                model_name=os.path.basename(export_model_path),
            )
            if export_output is None:
                # This should not happen
                raise RuntimeError("Failed to export model")

            model_fp32 = onnx.load(export_model_path)
            test_data = {input_names[i]: inp[i].float().cpu().numpy() for i in range(len(inp))}
            log.warning("Attempting to convert model in mixed precision, this may take a while...")
            with open(os.devnull, "w") as f, contextlib.redirect_stdout(f):
                # This function prints a lot of information that is not useful for the user
                model_fp16 = auto_convert_mixed_precision(
                    model_fp32, test_data, rtol=rtol, atol=atol, keep_io_types=False
                )
            onnx.save(model_fp16, export_model_path)

            onnx_model = onnx.load(export_model_path)
            # Check if ONNX model is valid
            onnx.checker.check_model(onnx_model)
            return True
        except Exception as e:
            raise RuntimeError(
                "Failed to export model with automatic mixed precision, check your model or disable ONNX export"
            ) from e
    else:
        log.info("Exported half precision ONNX model does not contain NaN values, model is stable")
        return True


def export_pytorch_model(model: nn.Module, output_path: str, model_name: str = "model.pth") -> str:
    """Export pytorch model's parameter dictionary using a deserialized state_dict.

    Args:
        model: PyTorch model to be exported
        output_path: Path to save the model
        model_name: Name of the exported model

    Returns:
        If the model is exported successfully, the path to the model is returned.

    """
    if isinstance(model, ModelSignatureWrapper):
        model = model.instance

    os.makedirs(output_path, exist_ok=True)
    model.eval()
    model.cpu()
    model_path = os.path.join(output_path, model_name)
    torch.save(model.state_dict(), model_path)
    log.info("Pytorch model saved to %s", os.path.join(output_path, model_name))

    return os.path.join(os.getcwd(), model_path)


def export_model(
    config: DictConfig,
    model: Any,
    export_folder: str,
    half_precision: bool,
    input_shapes: list[Any] | None = None,
    idx_to_class: dict[int, str] | None = None,
    pytorch_model_type: Literal["backbone", "model"] = "model",
    example_inputs: list[Any] | tuple[Any, ...] | torch.Tensor | None = None,
    rtol: float = 0.01,
    atol: float = 5e-3,
) -> tuple[dict[str, Any], dict[str, str]]:
    """Generate deployment models for the task.

    Args:
        config: Experiment config
        model: Model to be exported
        export_folder: Path to save the exported model
        half_precision: Whether to use half precision for the exported model
        input_shapes: Input shapes for the exported model
        idx_to_class: Mapping from class index to class name
        pytorch_model_type: Type of the pytorch model config to be exported, if it's backbone on disk we will save the
            config.backbone config, otherwise we will save the config.model
        example_inputs: If provided use this to evaluate the model instead of generating random inputs
        rtol: Relative tolerance for the ONNX safe export in fp16
        atol: Absolute tolerance for the ONNX safe export in fp16

    Returns:
        If the model is exported successfully, return a dictionary containing information about the exported model and
        a second dictionary containing the paths to the exported models. Otherwise, return two empty dictionaries.
    """
    if config.export is None or len(config.export.types) == 0:
        log.info("No export type specified skipping export")
        return {}, {}

    os.makedirs(export_folder, exist_ok=True)

    if input_shapes is None:
        # Try to get input shapes from config
        # If this is also None we will try to retrieve it from the ModelSignatureWrapper, if it fails we can't export
        input_shapes = config.export.input_shapes

    export_paths = {}

    for export_type in config.export.types:
        if export_type == "torchscript":
            out = export_torchscript_model(
                model=model,
                input_shapes=input_shapes,
                output_path=export_folder,
                half_precision=half_precision,
                example_inputs=example_inputs,
            )

            if out is None:
                log.warning("Torchscript export failed, enable debug logging for more details")
                continue

            export_path, input_shapes = out
            export_paths[export_type] = export_path
        elif export_type == "pytorch":
            export_path = export_pytorch_model(
                model=model,
                output_path=export_folder,
            )
            export_paths[export_type] = export_path
            with open(os.path.join(export_folder, "model_config.yaml"), "w") as f:
                OmegaConf.save(getattr(config, pytorch_model_type), f, resolve=True)
        elif export_type == "onnx":
            if not hasattr(config.export, "onnx"):
                log.warning("No onnx configuration found, skipping onnx export")
                continue

            out = export_onnx_model(
                model=model,
                output_path=export_folder,
                onnx_config=config.export.onnx,
                input_shapes=input_shapes,
                half_precision=half_precision,
                example_inputs=example_inputs,
                rtol=rtol,
                atol=atol,
            )

            if out is None:
                log.warning("ONNX export failed, enable debug logging for more details")
                continue

            export_path, input_shapes = out
            export_paths[export_type] = export_path
        else:
            log.warning("Export type: %s not implemented", export_type)

    if len(export_paths) == 0:
        log.warning("No export type was successful, no model will be available for deployment")
        return {}, export_paths

    model_json = {
        "input_size": input_shapes,
        "classes": idx_to_class,
        "mean": list(config.transforms.mean),
        "std": list(config.transforms.std),
    }

    return model_json, export_paths


def import_deployment_model(
    model_path: str,
    inference_config: DictConfig,
    device: str,
    model_architecture: nn.Module | None = None,
) -> BaseEvaluationModel:
    """Try to import a model for deployment, currently only supports torchscript .pt files and
    state dictionaries .pth files.

    Args:
        model_path: Path to the model
        inference_config: Inference configuration, should contain keys for the different deployment models
        device: Device to load the model on
        model_architecture: Optional model architecture to use for loading a plain pytorch model

    Returns:
        A tuple containing the model and the model type
    """
    log.info("Importing trained model")

    file_extension = os.path.splitext(os.path.basename(model_path))[1]
    deployment_model: BaseEvaluationModel | None = None

    if file_extension == ".pt":
        deployment_model = TorchscriptEvaluationModel(config=inference_config.torchscript)
    elif file_extension == ".pth":
        if model_architecture is None:
            raise ValueError("model_architecture must be specified when loading a .pth file")

        deployment_model = TorchEvaluationModel(config=inference_config.pytorch, model_architecture=model_architecture)
    elif file_extension == ".onnx":
        deployment_model = ONNXEvaluationModel(config=inference_config.onnx)

    if deployment_model is not None:
        deployment_model.load_from_disk(model_path=model_path, device=device)

        log.info("Imported %s model", deployment_model.__class__.__name__)

        return deployment_model

    raise ValueError(f"Unable to load model with extension {file_extension}, valid extensions are: ['.pt', 'pth']")


# This may be better as a dict?
def get_export_extension(export_type: str) -> str:
    """Get the extension of the exported model.

    Args:
        export_type: The type of the exported model.

    Returns:
        The extension of the exported model.
    """
    if export_type == "onnx":
        extension = "onnx"
    elif export_type == "torchscript":
        extension = "pt"
    elif export_type == "pytorch":
        extension = "pth"
    else:
        raise ValueError(f"Unsupported export type {export_type}")

    return extension



from __future__ import annotations

from collections.abc import Callable
from typing import Any

import pytorch_lightning as pl
import sklearn
import torch
import torchmetrics
from pytorch_lightning.utilities.types import OptimizerLRScheduler
from sklearn.linear_model import LogisticRegression
from torch import nn
from torch.optim import Optimizer

from quadra.models.base import ModelSignatureWrapper

__all__ = ["BaseLightningModule", "SSLModule"]


class BaseLightningModule(pl.LightningModule):
    """Base lightning module.

    Args:
        model: Network Module used for extract features
        optimizer: optimizer of the training. If None a default Adam is used.
        lr_scheduler: lr scheduler. If None a default ReduceLROnPlateau is used.
    """

    def __init__(
        self,
        model: nn.Module,
        optimizer: Optimizer | None = None,
        lr_scheduler: object | None = None,
        lr_scheduler_interval: str | None = "epoch",
    ):
        super().__init__()
        self.model = ModelSignatureWrapper(model)
        self.optimizer = optimizer
        self.schedulers = lr_scheduler
        self.lr_scheduler_interval = lr_scheduler_interval

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward method
        Args:
            x: input tensor.

        Returns:
            model inference
        """
        return self.model(x)

    def configure_optimizers(self) -> OptimizerLRScheduler:
        """Get default optimizer if not passed a value.

        Returns:
            optimizer and lr scheduler as Tuple containing a list of optimizers and a list of lr schedulers
        """
        # get default optimizer
        if getattr(self, "optimizer", None) is None or not self.optimizer:
            self.optimizer = torch.optim.Adam(self.model.parameters(), lr=3e-4)

        # get default scheduler
        if getattr(self, "schedulers", None) is None or not self.schedulers:
            self.schedulers = torch.optim.lr_scheduler.ReduceLROnPlateau(self.optimizer, factor=0.5, patience=30)

        lr_scheduler_conf = {
            "scheduler": self.schedulers,
            "interval": self.lr_scheduler_interval,
            "monitor": "val_loss",
            "strict": False,
        }
        return [self.optimizer], [lr_scheduler_conf]  # type: ignore[return-value]

    # pylint: disable=unused-argument
    def optimizer_zero_grad(self, epoch, batch_idx, optimizer, optimizer_idx: int = 0):
        """Redefine optimizer zero grad."""
        optimizer.zero_grad(set_to_none=True)


class SSLModule(BaseLightningModule):
    """Base module for self supervised learning.

    Args:
        model: Network Module used for extract features
        criterion: SSL loss to be applied
        classifier: Standard sklearn classifiers
        optimizer: optimizer of the training. If None a default Adam is used.
        lr_scheduler: lr scheduler. If None a default ReduceLROnPlateau is used.
    """

    def __init__(
        self,
        model: nn.Module,
        criterion: nn.Module,
        classifier: sklearn.base.ClassifierMixin | None = None,
        optimizer: Optimizer | None = None,
        lr_scheduler: object | None = None,
        lr_scheduler_interval: str | None = "epoch",
    ):
        super().__init__(model, optimizer, lr_scheduler, lr_scheduler_interval)
        self.criterion = criterion
        self.classifier_train_loader: torch.utils.data.DataLoader | None
        if classifier is None:
            self.classifier = LogisticRegression(max_iter=10000, n_jobs=8, random_state=42)
        else:
            self.classifier = classifier

        self.val_acc = torchmetrics.Accuracy()

    def fit_estimator(self):
        """Fit a classifier on the embeddings extracted from the current trained model."""
        targets = []
        train_embeddings = []
        self.model.eval()
        with torch.no_grad():
            for im, target in self.classifier_train_loader:
                emb = self.model(im.to(self.device))
                targets.append(target)
                train_embeddings.append(emb)
        targets = torch.cat(targets, dim=0).cpu().numpy()
        train_embeddings = torch.cat(train_embeddings, dim=0).cpu().numpy()
        self.classifier.fit(train_embeddings, targets)

    def calculate_accuracy(self, batch):
        """Calculate accuracy on a batch of data."""
        images, labels = batch
        with torch.no_grad():
            embedding = self.model(images).cpu().numpy()

        predictions = self.classifier.predict(embedding)
        labels = labels.detach()
        acc = self.val_acc(torch.tensor(predictions, device=self.device), labels)

        return acc

    # TODO: In multiprocessing mode, this function is called multiple times, how can we avoid this?
    def on_validation_start(self) -> None:
        if not hasattr(self, "classifier_train_loader") and hasattr(self.trainer, "datamodule"):
            self.classifier_train_loader = self.trainer.datamodule.classifier_train_dataloader()

        if self.classifier_train_loader is not None:
            self.fit_estimator()

    def validation_step(self, batch: tuple[tuple[torch.Tensor, torch.Tensor], torch.Tensor], batch_idx: int) -> None:
        # pylint: disable=unused-argument
        if self.classifier_train_loader is None:
            # Compute loss
            (im_x, im_y), _ = batch
            z1 = self(im_x)
            z2 = self(im_y)
            loss = self.criterion(z1, z2)

            self.log(
                "val_loss",
                loss,
                on_epoch=True,
                on_step=True,
                logger=True,
                prog_bar=True,
            )
            return loss

        acc = self.calculate_accuracy(batch)
        self.log("val_acc", acc, on_epoch=True, on_step=False, logger=True, prog_bar=True)
        return None


class SegmentationModel(BaseLightningModule):
    """Generic segmentation model.

    Args:
        model: segmentation model to be used.
        loss_fun: loss function to be used.
        optimizer: Optimizer to be used. Defaults to None.
        lr_scheduler: lr scheduler to be used. Defaults to None.
    """

    def __init__(
        self,
        model: torch.nn.Module,
        loss_fun: Callable,
        optimizer: Optimizer | None = None,
        lr_scheduler: object | None = None,
    ):
        super().__init__(model, optimizer, lr_scheduler)
        self.loss_fun = loss_fun

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward method
        Args:
            x: input tensor.

        Returns:
            model inference
        """
        x = self.model(x)
        return x

    def step(self, batch: tuple[torch.Tensor, torch.Tensor, torch.Tensor]) -> tuple[torch.Tensor, torch.Tensor]:
        """Compute loss
        Args:
            batch: batch.

        Returns:
            Prediction and target masks
        """
        images, target_masks, _ = batch
        pred_masks = self(images)
        if len(pred_masks.shape) == 3:
            pred_masks = pred_masks.unsqueeze(1)
        if len(target_masks.shape) == 3:
            target_masks = target_masks.unsqueeze(1)
        assert pred_masks.shape == target_masks.shape

        return pred_masks, target_masks

    def compute_loss(self, pred_masks: torch.Tensor, target_masks: torch.Tensor) -> torch.Tensor:
        """Compute loss
        Args:
            pred_masks: predicted masks
            target_masks: target masks.

        Returns:
            The computed loss

        """
        loss = self.loss_fun(pred_masks, target_masks)
        return loss

    def training_step(self, batch: tuple[torch.Tensor, torch.Tensor, torch.Tensor], batch_idx: int):
        """Training step."""
        # pylint: disable=unused-argument
        pred_masks, target_masks = self.step(batch)
        loss = self.compute_loss(pred_masks, target_masks)
        self.log_dict(
            {"loss": loss},
            on_step=True,
            on_epoch=True,
            prog_bar=True,
        )
        return loss

    def validation_step(self, batch: tuple[torch.Tensor, torch.Tensor, torch.Tensor], batch_idx):
        """Validation step."""
        # pylint: disable=unused-argument
        pred_masks, target_masks = self.step(batch)
        loss = self.compute_loss(pred_masks, target_masks)
        self.log_dict(
            {"val_loss": loss},
            on_step=True,
            on_epoch=True,
            prog_bar=True,
        )
        return loss

    def test_step(self, batch: tuple[torch.Tensor, torch.Tensor, torch.Tensor], batch_idx: int):
        """Test step."""
        # pylint: disable=unused-argument
        pred_masks, target_masks = self.step(batch)
        loss = self.compute_loss(pred_masks, target_masks)
        self.log_dict(
            {"test_loss": loss},
            on_step=True,
            on_epoch=True,
            prog_bar=True,
        )
        return loss

    def predict_step(
        self,
        batch: tuple[torch.Tensor, torch.Tensor, torch.Tensor],
        batch_idx: int,
        dataloader_idx: int | None = None,
    ) -> Any:
        """Predict step."""
        # pylint: disable=unused-argument
        images, masks, labels = batch
        pred_masks = self(images)
        return images.cpu(), masks.cpu(), pred_masks.cpu(), labels.cpu()


class SegmentationModelMulticlass(SegmentationModel):
    """Generic multiclass segmentation model.

    Args:
        model: segmentation model to be used.
        loss_fun: loss function to be used.
        optimizer: Optimizer to be used. Defaults to None.
        lr_scheduler: lr scheduler to be used. Defaults to None.
    """

    def __init__(
        self,
        model: torch.nn.Module,
        loss_fun: Callable,
        optimizer: Optimizer | None = None,
        lr_scheduler: object | None = None,
    ):
        super().__init__(model=model, optimizer=optimizer, lr_scheduler=lr_scheduler, loss_fun=loss_fun)

    def step(self, batch: tuple[torch.Tensor, torch.Tensor, torch.Tensor]) -> tuple[torch.Tensor, torch.Tensor]:
        """Compute step
        Args:
            batch: batch.

        Returns:
            prediction, target

        """
        images, target_masks, _ = batch
        pred_masks = self(images)

        return pred_masks, target_masks



from __future__ import annotations

from typing import Any

from hydra.core.hydra_config import HydraConfig
from omegaconf import OmegaConf


def multirun_subdir_beautify(subdir: str) -> str:
    """Change the subdir name to be more readable and usable, this function will replace / with | to avoid creating
    undesired subdirectories and remove the left part of the equals sign to avoid having too long names.

    Args:
        subdir: The subdir name.

    Returns:
        The beautified subdir name.

    Examples:
        >>> multirun_subdir_beautify("experiment=pippo/anomaly/padim,trainer.batch_size=32")
        "pippo|anomaly|padim,32"
    """
    hydra_cfg = HydraConfig.get()
    if hydra_cfg.mode is None or hydra_cfg.mode.name == "RUN":
        return subdir
    # Remove slashes to avoid creating multiple subdirs
    # TODO: if right side of the equals sign has `,` this will not work.
    subdir_list = subdir.replace("/", "|").split(",")
    subdir = ",".join([x.split("=")[1].replace(" ", "") for x in subdir_list])

    return subdir


def as_tuple(*args: Any) -> tuple[Any, ...]:
    """Resolves a list of arguments to a tuple."""
    return tuple(args)


def register_resolvers() -> None:
    """Register custom resolver."""
    OmegaConf.register_new_resolver("multirun_subdir_beautify", multirun_subdir_beautify)
    OmegaConf.register_new_resolver("as_tuple", as_tuple)



from __future__ import annotations

from typing import Any

import timm
import torch
from timm.models.helpers import load_checkpoint
from torch import nn
from torchvision import models

from quadra.models.classification.base import BaseNetworkBuilder
from quadra.utils.logger import get_logger

log = get_logger(__name__)


class TorchHubNetworkBuilder(BaseNetworkBuilder):
    """TorchHub feature extractor, with the possibility to map features to an hypersphere.

    Args:
        repo_or_dir: The name of the repository or the path to the directory containing the model.
        model_name: The name of the model within the repository.
        pretrained: Whether to load the pretrained weights for the model.
        pre_classifier: Pre classifier as a torch.nn.Module. Defaults to nn.Identity() if None.
        classifier: Classifier as a torch.nn.Module. Defaults to nn.Identity() if None.
        freeze: Whether to freeze the feature extractor. Defaults to True.
        hyperspherical: Whether to map features to an hypersphere. Defaults to False.
        flatten_features: Whether to flatten the features before the pre_classifier. Defaults to True.
        checkpoint_path: Path to a checkpoint to load after the model is initialized. Defaults to None.
        **torch_hub_kwargs: Additional arguments to pass to torch.hub.load
    """

    def __init__(
        self,
        repo_or_dir: str,
        model_name: str,
        pretrained: bool = True,
        pre_classifier: nn.Module | None = None,
        classifier: nn.Module | None = None,
        freeze: bool = True,
        hyperspherical: bool = False,
        flatten_features: bool = True,
        checkpoint_path: str | None = None,
        **torch_hub_kwargs: Any,
    ):
        self.pretrained = pretrained
        features_extractor = torch.hub.load(
            repo_or_dir=repo_or_dir, model=model_name, pretrained=self.pretrained, **torch_hub_kwargs
        )
        if checkpoint_path:
            log.info("Loading checkpoint from %s", checkpoint_path)
            load_checkpoint(model=features_extractor, checkpoint_path=checkpoint_path)

        super().__init__(
            features_extractor=features_extractor,
            pre_classifier=pre_classifier,
            classifier=classifier,
            freeze=freeze,
            hyperspherical=hyperspherical,
            flatten_features=flatten_features,
        )


class TorchVisionNetworkBuilder(BaseNetworkBuilder):
    """Torchvision feature extractor, with the possibility to map features to an hypersphere.

    Args:
        model_name: Torchvision model function that will be evaluated, for example: torchvision.models.resnet18.
        pretrained: Whether to load the pretrained weights for the model.
        pre_classifier: Pre classifier as a torch.nn.Module. Defaults to nn.Identity() if None.
        classifier: Classifier as a torch.nn.Module. Defaults to nn.Identity() if None.
        freeze: Whether to freeze the feature extractor. Defaults to True.
        hyperspherical: Whether to map features to an hypersphere. Defaults to False.
        flatten_features: Whether to flatten the features before the pre_classifier. Defaults to True.
        checkpoint_path: Path to a checkpoint to load after the model is initialized. Defaults to None.
        **torchvision_kwargs: Additional arguments to pass to the model function.
    """

    def __init__(
        self,
        model_name: str,
        pretrained: bool = True,
        pre_classifier: nn.Module | None = None,
        classifier: nn.Module | None = None,
        freeze: bool = True,
        hyperspherical: bool = False,
        flatten_features: bool = True,
        checkpoint_path: str | None = None,
        **torchvision_kwargs: Any,
    ):
        self.pretrained = pretrained
        model_function = models.__dict__[model_name]
        features_extractor = model_function(pretrained=self.pretrained, progress=True, **torchvision_kwargs)
        if checkpoint_path:
            log.info("Loading checkpoint from %s", checkpoint_path)
            load_checkpoint(model=features_extractor, checkpoint_path=checkpoint_path)

        # Remove classifier
        features_extractor.classifier = nn.Identity()
        super().__init__(
            features_extractor=features_extractor,
            pre_classifier=pre_classifier,
            classifier=classifier,
            freeze=freeze,
            hyperspherical=hyperspherical,
            flatten_features=flatten_features,
        )


class TimmNetworkBuilder(BaseNetworkBuilder):
    """Torchvision feature extractor, with the possibility to map features to an hypersphere.

    Args:
        model_name: Timm model name
        pretrained: Whether to load the pretrained weights for the model.
        pre_classifier: Pre classifier as a torch.nn.Module. Defaults to nn.Identity() if None.
        classifier: Classifier as a torch.nn.Module. Defaults to nn.Identity() if None.
        freeze: Whether to freeze the feature extractor. Defaults to True.
        hyperspherical: Whether to map features to an hypersphere. Defaults to False.
        flatten_features: Whether to flatten the features before the pre_classifier. Defaults to True.
        checkpoint_path: Path to a checkpoint to load after the model is initialized. Defaults to None.
        **timm_kwargs: Additional arguments to pass to timm.create_model
    """

    def __init__(
        self,
        model_name: str,
        pretrained: bool = True,
        pre_classifier: nn.Module | None = None,
        classifier: nn.Module | None = None,
        freeze: bool = True,
        hyperspherical: bool = False,
        flatten_features: bool = True,
        checkpoint_path: str | None = None,
        **timm_kwargs: Any,
    ):
        self.pretrained = pretrained
        features_extractor = timm.create_model(
            model_name, pretrained=self.pretrained, num_classes=0, checkpoint_path=checkpoint_path, **timm_kwargs
        )

        super().__init__(
            features_extractor=features_extractor,
            pre_classifier=pre_classifier,
            classifier=classifier,
            freeze=freeze,
            hyperspherical=hyperspherical,
            flatten_features=flatten_features,
        )



from __future__ import annotations

import warnings

import torch
import torch.nn.functional as F
from torch import nn


def one_hot(
    labels: torch.Tensor,
    num_classes: int,
    device: torch.device | None = None,
    dtype: torch.dtype | None = None,
    eps: float = 1e-6,
) -> torch.Tensor:
    r"""Convert an integer label x-D tensor to a one-hot (x+1)-D tensor.

    Args:
        labels: tensor with labels of shape :math:`(N, *)`, where N is batch size.
            Each value is an integer representing correct classification.
        num_classes: number of classes in labels.
        device: the desired device of returned tensor.
        dtype: the desired data type of returned tensor.
        eps: a value added to the returned tensor.

    Returns:
        the labels in one hot tensor of shape :math:`(N, C, *)`,

    Examples:
        >>> labels = torch.LongTensor([[[0, 1], [2, 0]]])
        >>> one_hot(labels, num_classes=3)
        tensor([[[[1.0000e+00, 1.0000e-06],
                  [1.0000e-06, 1.0000e+00]],
        <BLANKLINE>
                 [[1.0000e-06, 1.0000e+00],
                  [1.0000e-06, 1.0000e-06]],
        <BLANKLINE>
                 [[1.0000e-06, 1.0000e-06],
                  [1.0000e+00, 1.0000e-06]]]])

    """
    if not isinstance(labels, torch.Tensor):
        raise TypeError(f"Input labels type is not a torch.Tensor. Got {type(labels)}")

    if not labels.dtype == torch.int64:
        raise ValueError(f"labels must be of the same dtype torch.int64. Got: {labels.dtype}")

    if num_classes < 1:
        raise ValueError(f"The number of classes must be bigger than one. Got: {num_classes}")

    shape = labels.shape
    one_hot_output = torch.zeros((shape[0], num_classes) + shape[1:], device=device, dtype=dtype)

    return one_hot_output.scatter_(1, labels.unsqueeze(1), 1.0) + eps


# based on: # https://github.com/zhezh/focalloss/blob/master/focalloss.py
def focal_loss(
    input_tensor: torch.Tensor,
    target: torch.Tensor,
    alpha: float,
    gamma: float = 2.0,
    reduction: str = "none",
    eps: float | None = None,
) -> torch.Tensor:
    r"""Criterion that computes Focal loss.

    According to :cite:`lin2018focal`, the Focal loss is computed as follows:

    .. math::

        \text{FL}(p_t) = -\alpha_t (1 - p_t)^{\gamma} \, \text{log}(p_t)

    Where:
       - :math:`p_t` is the model's estimated probability for each class.

    Args:
        input_tensor: Logits tensor with shape :math:`(N, C, *)` where C = number of classes.
        target: Labels tensor with shape :math:`(N, *)` where each value is :math:`0 ≤ targets[i] ≤ C−1`.
        alpha: Weighting factor :math:`\alpha \in [0, 1]`.
        gamma: Focusing parameter :math:`\gamma >= 0`.
        reduction: Specifies the reduction to apply to the
            output: ``'none'`` | ``'mean'`` | ``'sum'``. ``'none'``: no reduction
            will be applied, ``'mean'``: the sum of the output will be divided by
            the number of elements in the output, ``'sum'``: the output will be
            summed.
        eps: Deprecated: scalar to enforce numerical stabiliy. This is no longer used.

    Returns:
        The computed loss.

    Example:
        >>> N = 5  # num_classes
        >>> input = torch.randn(1, N, 3, 5, requires_grad=True)
        >>> target = torch.empty(1, 3, 5, dtype=torch.long).random_(N)
        >>> output = focal_loss(input, target, alpha=0.5, gamma=2.0, reduction='mean')
        >>> output.backward()
    """
    if eps is not None and not torch.jit.is_scripting():
        warnings.warn(
            "`focal_loss` has been reworked for improved numerical stability "
            "and the `eps` argument is no longer necessary",
            DeprecationWarning,
            stacklevel=2,
        )

    if not isinstance(input_tensor, torch.Tensor):
        raise TypeError(f"Input type is not a torch.Tensor. Got {type(input_tensor)}")

    if not len(input_tensor.shape) >= 2:
        raise ValueError(f"Invalid input shape, we expect BxCx*. Got: {input_tensor.shape}")

    if input_tensor.size(0) != target.size(0):
        raise ValueError(
            f"Expected input batch_size ({input_tensor.size(0)}) to match target batch_size ({target.size(0)})."
        )

    n = input_tensor.size(0)
    out_size = (n,) + input_tensor.size()[2:]
    if target.size()[1:] != input_tensor.size()[2:]:
        raise ValueError(f"Expected target size {out_size}, got {target.size()}")

    if not input_tensor.device == target.device:
        raise ValueError(f"input and target must be in the same device. Got: {input_tensor.device} and {target.device}")

    # compute softmax over the classes axis
    input_soft: torch.Tensor = F.softmax(input_tensor, dim=1)
    log_input_soft: torch.Tensor = F.log_softmax(input_tensor, dim=1)

    # create the labels one hot tensor
    target_one_hot: torch.Tensor = one_hot(
        target, num_classes=input_tensor.shape[1], device=input_tensor.device, dtype=input_tensor.dtype
    )

    # compute the actual focal loss
    weight = torch.pow(-input_soft + 1.0, gamma)

    focal = -alpha * weight * log_input_soft
    loss_tmp = torch.einsum("bc...,bc...->b...", (target_one_hot, focal))

    if reduction == "none":
        loss = loss_tmp
    elif reduction == "mean":
        loss = torch.mean(loss_tmp)
    elif reduction == "sum":
        loss = torch.sum(loss_tmp)
    else:
        raise NotImplementedError(f"Invalid reduction mode: {reduction}")
    return loss


class FocalLoss(nn.Module):
    r"""Criterion that computes Focal loss.

    According to :cite:`lin2018focal`, the Focal loss is computed as follows:

    .. math::

        \text{FL}(p_t) = -\alpha_t (1 - p_t)^{\gamma} \, \text{log}(p_t)

    Where:
       - :math:`p_t` is the model's estimated probability for each class.

    Args:
        alpha: Weighting factor :math:`\alpha \in [0, 1]`.
        gamma: Focusing parameter :math:`\gamma >= 0`.
        reduction: Specifies the reduction to apply to the
            output: ``'none'`` | ``'mean'`` | ``'sum'``. ``'none'``: no reduction
            will be applied, ``'mean'``: the sum of the output will be divided by
            the number of elements in the output, ``'sum'``: the output will be
            summed.
        eps: Deprecated: scalar to enforce numerical stability. This is no longer
            used.

    Shape:
        - Input: :math:`(N, C, *)` where C = number of classes.
        - Target: :math:`(N, *)` where each value is
          :math:`0 ≤ targets[i] ≤ C−1`.

    Example:
        >>> N = 5  # num_classes
        >>> kwargs = {"alpha": 0.5, "gamma": 2.0, "reduction": 'mean'}
        >>> criterion = FocalLoss(**kwargs)
        >>> input = torch.randn(1, N, 3, 5, requires_grad=True)
        >>> target = torch.empty(1, 3, 5, dtype=torch.long).random_(N)
        >>> output = criterion(input, target)
        >>> output.backward()
    """

    def __init__(self, alpha: float, gamma: float = 2.0, reduction: str = "none", eps: float | None = None) -> None:
        super().__init__()
        self.alpha: float = alpha
        self.gamma: float = gamma
        self.reduction: str = reduction
        self.eps: float | None = eps

    def forward(self, input_tensor: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        """Forward call computation."""
        return focal_loss(input_tensor, target, self.alpha, self.gamma, self.reduction, self.eps)


def binary_focal_loss_with_logits(
    input_tensor: torch.Tensor,
    target: torch.Tensor,
    alpha: float = 0.25,
    gamma: float = 2.0,
    reduction: str = "none",
    eps: float | None = None,
) -> torch.Tensor:
    r"""Function that computes Binary Focal loss.

    .. math::

        \text{FL}(p_t) = -\alpha_t (1 - p_t)^{\gamma} \, \text{log}(p_t)

    where:
       - :math:`p_t` is the model's estimated probability for each class.

    Args:
        input_tensor: input data tensor of arbitrary shape.
        target: the target tensor with shape matching input.
        alpha: Weighting factor for the rare class :math:`\alpha \in [0, 1]`.
        gamma: Focusing parameter :math:`\gamma >= 0`.
        reduction: Specifies the reduction to apply to the
            output: ``'none'`` | ``'mean'`` | ``'sum'``. ``'none'``: no reduction
            will be applied, ``'mean'``: the sum of the output will be divided by
            the number of elements in the output, ``'sum'``: the output will be
            summed.
        eps: Deprecated: scalar for numerically stability when dividing. This is no longer used.

    Returns:
        the computed loss.

    Examples:
        >>> kwargs = {"alpha": 0.25, "gamma": 2.0, "reduction": 'mean'}
        >>> logits = torch.tensor([[[6.325]],[[5.26]],[[87.49]]])
        >>> labels = torch.tensor([[[1.]],[[1.]],[[0.]]])
        >>> binary_focal_loss_with_logits(logits, labels, **kwargs)
        tensor(21.8725)
    """
    if eps is not None and not torch.jit.is_scripting():
        warnings.warn(
            "`binary_focal_loss_with_logits` has been reworked for improved numerical stability "
            "and the `eps` argument is no longer necessary",
            DeprecationWarning,
            stacklevel=2,
        )

    if not isinstance(input_tensor, torch.Tensor):
        raise TypeError(f"Input type is not a torch.Tensor. Got {type(input_tensor)}")

    if not len(input_tensor.shape) >= 2:
        raise ValueError(f"Invalid input shape, we expect BxCx*. Got: {input_tensor.shape}")

    if input_tensor.size(0) != target.size(0):
        raise ValueError(
            f"Expected input batch_size ({input_tensor.size(0)}) to match target batch_size ({target.size(0)})."
        )

    probs_pos = torch.sigmoid(input_tensor)
    probs_neg = torch.sigmoid(-input_tensor)
    loss_tmp = -alpha * torch.pow(probs_neg, gamma) * target * F.logsigmoid(input_tensor) - (1 - alpha) * torch.pow(
        probs_pos, gamma
    ) * (1.0 - target) * F.logsigmoid(-input_tensor)

    if reduction == "none":
        loss = loss_tmp
    elif reduction == "mean":
        loss = torch.mean(loss_tmp)
    elif reduction == "sum":
        loss = torch.sum(loss_tmp)
    else:
        raise NotImplementedError(f"Invalid reduction mode: {reduction}")
    return loss


class BinaryFocalLossWithLogits(nn.Module):
    r"""Criterion that computes Focal loss.

    According to :cite:`lin2018focal`, the Focal loss is computed as follows:

    .. math::

        \text{FL}(p_t) = -\alpha_t (1 - p_t)^{\gamma} \, \text{log}(p_t)

    where:
       - :math:`p_t` is the model's estimated probability for each class.

    Args:
        alpha: Weighting factor for the rare class :math:`\alpha \in [0, 1]`.
        gamma: Focusing parameter :math:`\gamma >= 0`.
        reduction: Specifies the reduction to apply to the
            output: ``'none'`` | ``'mean'`` | ``'sum'``. ``'none'``: no reduction
            will be applied, ``'mean'``: the sum of the output will be divided by
            the number of elements in the output, ``'sum'``: the output will be
            summed.

    Shape:
        - Input: :math:`(N, *)`.
        - Target: :math:`(N, *)`.

    Examples:
        >>> kwargs = {"alpha": 0.25, "gamma": 2.0, "reduction": 'mean'}
        >>> loss = BinaryFocalLossWithLogits(**kwargs)
        >>> input = torch.randn(1, 3, 5, requires_grad=True)
        >>> target = torch.empty(1, 3, 5, dtype=torch.long).random_(2)
        >>> output = loss(input, target)
        >>> output.backward()
    """

    def __init__(self, alpha: float, gamma: float = 2.0, reduction: str = "none") -> None:
        super().__init__()
        self.alpha: float = alpha
        self.gamma: float = gamma
        self.reduction: str = reduction

    def forward(self, input_tensor: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        """Forward call computation."""
        return binary_focal_loss_with_logits(input_tensor, target, self.alpha, self.gamma, self.reduction)



from __future__ import annotations

import os
from collections.abc import Callable
from typing import Any

import albumentations
import cv2
import numpy as np
import torch

from quadra.utils.deprecation import deprecated
from quadra.utils.imaging import keep_aspect_ratio_resize
from quadra.utils.segmentation import smooth_mask


# DEPRECATED -> we can use SegmentationDatasetMulticlass also for one class segmentation
@deprecated("Use SegmentationDatasetMulticlass instead")
class SegmentationDataset(torch.utils.data.Dataset):
    """Custom SegmentationDataset class for loading images and masks.

    Args:
        image_paths: List of paths to images.
        mask_paths: List of paths to masks.
        batch_size: Batch size.
        object_masks: List of paths to object masks.
        resize: Resize image to this size.
        mask_preprocess: Preprocess mask.
        labels: List of labels.
        transform: Transformations to apply to images and masks.
        mask_smoothing: Smooth mask.
        defect_transform: Transformations to apply to images and masks for defects.
    """

    def __init__(
        self,
        image_paths: list[str],
        mask_paths: list[str],
        batch_size: int | None = None,
        object_masks: list[np.ndarray | Any] | None = None,
        resize: int = 224,
        mask_preprocess: Callable | None = None,
        labels: list[str] | None = None,
        transform: albumentations.Compose | None = None,
        mask_smoothing: bool = False,
        defect_transform: albumentations.Compose | None = None,
    ):
        self.transform = transform
        self.defect_transform = defect_transform
        self.image_paths = image_paths
        self.mask_paths = mask_paths
        self.labels = labels
        self.mask_preprocess = mask_preprocess
        self.resize = resize
        self.object_masks = object_masks
        self.data_len = len(self.image_paths)
        self.batch_size = None if batch_size is None else max(batch_size, self.data_len)
        self.smooth_mask = mask_smoothing

    def __getitem__(self, index):
        # This is required to avoid infinite loop when running the dataset outside of a dataloader
        if self.batch_size is not None and self.batch_size == index:
            raise StopIteration

        if self.batch_size is None and self.data_len == index:
            raise StopIteration

        index = index % self.data_len
        image_path = self.image_paths[index]

        image = cv2.imread(str(image_path))
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        object_mask_path = self.object_masks[index] if self.object_masks is not None else None
        if object_mask_path is not None:
            object_mask = cv2.imread(str(object_mask_path), 0) if os.path.isfile(object_mask_path) else None
        else:
            object_mask = None
        label = self.labels[index] if self.labels is not None else None
        if (
            self.mask_paths[index] is np.nan
            or self.mask_paths[index] is None
            or not os.path.isfile(self.mask_paths[index])
        ):
            mask = np.zeros((image.shape[0], image.shape[1]), dtype=np.uint8)
        else:
            mask_path = self.mask_paths[index]
            mask = cv2.imread(str(mask_path), 0)
        if self.defect_transform is not None and label == 1 and np.sum(mask) == 0:
            if object_mask is not None:
                object_mask *= 255
            aug = self.defect_transform(image=image, mask=mask, object_mask=object_mask, label=label)
            image = aug["image"]
            mask = aug["mask"]
            label = aug["label"]
        if self.mask_preprocess:
            mask = self.mask_preprocess(mask)
            if object_mask is not None:
                object_mask = self.mask_preprocess(object_mask)
        if self.resize:
            image = keep_aspect_ratio_resize(image, self.resize)
            mask = keep_aspect_ratio_resize(mask, self.resize)
            if object_mask is not None:
                object_mask = keep_aspect_ratio_resize(object_mask, self.resize)

        if self.transform is not None:
            aug = self.transform(image=image, mask=mask)
            image = aug["image"]
            mask = aug["mask"]
        if isinstance(mask, np.ndarray):
            mask_sum = np.sum(mask)
        elif isinstance(mask, torch.Tensor):
            mask_sum = torch.sum(mask)
        else:
            raise ValueError("Unsupported type for mask")
        if mask_sum > 0 and (label is None or label == 0):
            label = 1
        if mask_sum == 0:
            label = 0

        if isinstance(image, np.ndarray):
            mask = (mask > 0).astype(np.uint8)

            if self.smooth_mask:
                mask = smooth_mask(mask)
            mask = np.expand_dims(mask, axis=0)
        else:
            mask = (mask > 0).int()
            if self.smooth_mask:
                mask = torch.from_numpy(smooth_mask(mask.numpy()))
            mask = mask.unsqueeze(0)

        return image, mask, label

    def __len__(self):
        if self.batch_size is None:
            return self.data_len

        return max(self.data_len, self.batch_size)


class SegmentationDatasetMulticlass(torch.utils.data.Dataset):
    """Custom SegmentationDataset class for loading images and multilabel masks.

    Args:
        image_paths: List of paths to images.
        mask_paths: List of paths to masks.
        idx_to_class: dict with corrispondence btw mask index and classes: {1: class_1, 2: class_2, ..., N: class_N}
        batch_size: Batch size.
        transform: Transformations to apply to images and masks.
        one_hot: if True return a binary mask (n_classxHxW), otherwise the labelled mask HxW. SMP loss requires the
            second format.
    """

    def __init__(
        self,
        image_paths: list[str],
        mask_paths: list[str],
        idx_to_class: dict,
        batch_size: int | None = None,
        transform: albumentations.Compose | None = None,
        one_hot: bool = False,
    ):
        self.transform = transform
        self.image_paths = image_paths
        self.mask_paths = mask_paths
        self.idx_to_class = idx_to_class
        self.data_len = len(self.image_paths)
        self.batch_size = None if batch_size is None else max(batch_size, self.data_len)
        self.one_hot = one_hot

    def _preprocess_mask(self, mask: np.ndarray):
        """Function to preprocess the mask -> needed for albumentations
        Args:
            mask: a numpy array of dimension HxW with values in [0] + self.idx_to_class.

        Output:
            a binary numpy array with dims len(self.idx_to_class) + 1 x H x W
        """
        multilayer_mask = np.zeros((len(self.idx_to_class) + 1, *mask.shape[:2]))
        # provide background information for completeness
        # single channel mask does not use it anyway.
        multilayer_mask[0] = (mask == 0).astype(np.uint8)
        for idx in self.idx_to_class:
            multilayer_mask[int(idx)] = (mask == int(idx)).astype(np.uint8)

        return multilayer_mask

    def __getitem__(self, index):
        """Get image and mask."""
        # This is required to avoid infinite loop when running the dataset outside of a dataloader
        if self.batch_size is not None and self.batch_size == index:
            raise StopIteration
        if self.batch_size is None and self.data_len == index:
            raise StopIteration

        index = index % self.data_len
        image_path = self.image_paths[index]

        image = cv2.imread(str(image_path))
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        if (
            self.mask_paths[index] is np.nan
            or self.mask_paths[index] is None
            or not os.path.isfile(self.mask_paths[index])
        ):
            mask = np.zeros((image.shape[0], image.shape[1]), dtype=np.uint8)
        else:
            mask_path = self.mask_paths[index]
            mask = cv2.imread(str(mask_path), 0)

        # we go back to binary masks avoid transformation errors
        mask = self._preprocess_mask(mask)

        if self.transform is not None:
            masks = list(mask)
            aug = self.transform(image=image, masks=masks)
            image = aug["image"]
            mask = np.stack(aug["masks"])  # C x H x W

        # we compute single channel mask again
        # zero is the background
        if not self.one_hot:  # one hot is done by smp dice loss
            mask_out = np.zeros(mask.shape[1:])
            for i in range(1, mask.shape[0]):
                mask_out[mask[i] == 1] = i
            # mask_out shape -> HxW
        else:
            mask_out = mask
            # mask_out shape -> CxHxW where C is number of classes (included the background)

        return image, mask_out.astype(int), 0

    def __len__(self):
        """Returns the dataset lenght."""
        if self.batch_size is None:
            return self.data_len

        return max(self.data_len, self.batch_size)



from __future__ import annotations

import torch
from torch import nn

from quadra.utils.models import trunc_normal_


class ProjectionHead(torch.nn.Module):
    """Base class for all projection and prediction heads.

    Args:
        blocks:
            List of tuples, each denoting one block of the projection head MLP.
            Each tuple reads (linear_layer, batch_norm_layer, non_linearity_layer).
            `batch_norm` layer can be possibly None, the same happens for
            `non_linearity_layer`.
    """

    def __init__(self, blocks: list[tuple[torch.nn.Module | None, ...]]):
        super().__init__()

        layers: list[nn.Module] = []
        for linear, batch_norm, non_linearity in blocks:
            if linear:
                layers.append(linear)
            if batch_norm:
                layers.append(batch_norm)
            if non_linearity:
                layers.append(non_linearity)
        self.layers = torch.nn.Sequential(*layers)

    def forward(self, x: torch.Tensor):
        return self.layers(x)


class ExpanderReducer(ProjectionHead):
    """Expander followed by a reducer."""

    def __init__(self, input_dim: int, hidden_dim: int, output_dim: int):
        super().__init__(
            [
                (
                    torch.nn.Linear(input_dim, hidden_dim, bias=False),
                    torch.nn.BatchNorm1d(hidden_dim),
                    torch.nn.ReLU(inplace=True),
                ),
                (
                    torch.nn.Linear(hidden_dim, output_dim, bias=False),
                    torch.nn.BatchNorm1d(output_dim, affine=False),
                    torch.nn.ReLU(inplace=True),
                ),
            ]
        )


class BarlowTwinsProjectionHead(ProjectionHead):
    """Projection head used for Barlow Twins.
    "The projector network has three linear layers, each with 8192 output
    units. The first two layers of the projector are followed by a batch
    normalization layer and rectified linear units." https://arxiv.org/abs/2103.03230.
    """

    def __init__(self, input_dim: int, hidden_dim: int, output_dim: int):
        super().__init__(
            [
                (
                    torch.nn.Linear(input_dim, hidden_dim, bias=False),
                    torch.nn.BatchNorm1d(hidden_dim),
                    torch.nn.ReLU(inplace=True),
                ),
                (
                    torch.nn.Linear(hidden_dim, hidden_dim, bias=False),
                    torch.nn.BatchNorm1d(hidden_dim),
                    torch.nn.ReLU(inplace=True),
                ),
                (torch.nn.Linear(hidden_dim, output_dim, bias=False), None, None),
            ]
        )


class SimCLRProjectionHead(ProjectionHead):
    """Projection head used for SimCLR.
    "We use a MLP with one hidden layer to obtain zi = g(h) = W_2 * σ(W_1 * h)
    where σ is a ReLU non-linearity." https://arxiv.org/abs/2002.05709.
    """

    def __init__(self, input_dim: int, hidden_dim: int, output_dim: int):
        super().__init__(
            [
                (
                    torch.nn.Linear(input_dim, hidden_dim),
                    None,
                    torch.nn.ReLU(inplace=True),
                ),
                (torch.nn.Linear(hidden_dim, output_dim), None, None),
            ]
        )


class SimCLRPredictionHead(ProjectionHead):
    """Prediction head used for SimCLR.
    "We set g(h) = W(2)σ(W(1)h), with the same input and output dimensionality (i.e. 2048)."
    https://arxiv.org/abs/2002.05709.
    """

    def __init__(self, input_dim: int, hidden_dim: int, output_dim: int):
        super().__init__(
            [
                (
                    torch.nn.Linear(input_dim, hidden_dim, bias=False),
                    torch.nn.BatchNorm1d(hidden_dim),
                    torch.nn.ReLU(inplace=True),
                ),
                (torch.nn.Linear(hidden_dim, output_dim, bias=False), None, None),
            ]
        )


class SimSiamProjectionHead(ProjectionHead):
    """Projection head used for SimSiam.
    "The projection MLP (in f) has BN applied to each fully-connected (fc)
    layer, including its output fc. Its output fc has no ReLU. The hidden fc is
    2048-d. This MLP has 3 layers." https://arxiv.org/abs/2011.10566.
    """

    def __init__(self, input_dim: int, hidden_dim: int, output_dim: int):
        super().__init__(
            [
                (
                    torch.nn.Linear(input_dim, hidden_dim, bias=False),
                    torch.nn.BatchNorm1d(hidden_dim),
                    torch.nn.ReLU(inplace=True),
                ),
                (
                    torch.nn.Linear(hidden_dim, hidden_dim, bias=False),
                    torch.nn.BatchNorm1d(hidden_dim, affine=False),
                    torch.nn.ReLU(inplace=True),
                ),
                (
                    torch.nn.Linear(hidden_dim, output_dim, bias=False),
                    torch.nn.BatchNorm1d(output_dim, affine=False),
                    None,
                ),
            ]
        )


class SimSiamPredictionHead(ProjectionHead):
    """Prediction head used for SimSiam.
    "The prediction MLP (h) has BN applied to its hidden fc layers. Its output
    fc does not have BN (...) or ReLU. This MLP has 2 layers." https://arxiv.org/abs/2011.10566.
    """

    def __init__(self, input_dim: int, hidden_dim: int, output_dim: int):
        super().__init__(
            [
                (
                    torch.nn.Linear(input_dim, hidden_dim, bias=False),
                    torch.nn.BatchNorm1d(hidden_dim),
                    torch.nn.ReLU(inplace=True),
                ),
                (torch.nn.Linear(hidden_dim, output_dim, bias=False), None, None),
            ]
        )


class BYOLPredictionHead(ProjectionHead):
    """Prediction head used for BYOL."""

    def __init__(self, input_dim: int, hidden_dim: int, output_dim: int):
        super().__init__(
            [
                (
                    torch.nn.Linear(input_dim, hidden_dim, bias=False),
                    torch.nn.BatchNorm1d(hidden_dim),
                    torch.nn.ReLU(inplace=True),
                ),
                (torch.nn.Linear(hidden_dim, output_dim, bias=False), None, None),
            ]
        )


class BYOLProjectionHead(ProjectionHead):
    """Projection head used for BYOL."""

    def __init__(self, input_dim: int, hidden_dim: int, output_dim: int):
        super().__init__(
            [
                (
                    torch.nn.Linear(input_dim, hidden_dim, bias=False),
                    torch.nn.BatchNorm1d(hidden_dim),
                    torch.nn.ReLU(inplace=True),
                ),
                (torch.nn.Linear(hidden_dim, output_dim, bias=False), None, None),
            ]
        )


class DinoProjectionHead(nn.Module):
    """Projection head used for Dino. This projection head does not have
    a batch norm layer.

    Args:
        input_dim: Input dimension for MLP head.
        output_dim: Output dimension (projection dimension) for MLP head.
        hidden_dim: Hidden dimension. Defaults to 512.
        bottleneck_dim: Bottleneck dimension. Defaults to 256.
        num_layers: Number of hidden layers used in MLP. Defaults to 3.
        norm_last_layer: Decides applying normalization before last layer.
            Defaults to False.
    """

    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        hidden_dim: int,
        use_bn: bool = False,
        norm_last_layer: bool = True,
        num_layers: int = 3,
        bottleneck_dim: int = 256,
    ):
        super().__init__()
        num_layers = max(num_layers, 1)
        self.mlp: nn.Linear | nn.Sequential
        if num_layers == 1:
            self.mlp = nn.Linear(input_dim, bottleneck_dim)
        else:
            layers: list[nn.Module] = [nn.Linear(input_dim, hidden_dim)]
            if use_bn:
                layers.append(nn.BatchNorm1d(hidden_dim))
            layers.append(nn.GELU())
            for _ in range(num_layers - 2):
                layers.append(nn.Linear(hidden_dim, hidden_dim))
                if use_bn:
                    layers.append(nn.BatchNorm1d(hidden_dim))
                layers.append(nn.GELU())
            layers.append(nn.Linear(hidden_dim, bottleneck_dim))
            self.mlp = nn.Sequential(*layers)
        self.apply(self._init_weights)
        self.last_layer = nn.utils.weight_norm(nn.Linear(bottleneck_dim, output_dim, bias=False))
        self.last_layer.weight_g.data.fill_(1)
        if norm_last_layer:
            self.last_layer.weight_g.requires_grad = False

    def _init_weights(self, m):
        """Initialize the weights of the projection head."""
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=0.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)

    def forward(self, x):
        x = self.mlp(x)
        x = nn.functional.normalize(x, dim=-1, p=2)
        x = self.last_layer(x)
        return x


class MultiCropModel(nn.Module):
    """MultiCrop model for DINO augmentation.

    It takes 2 global crops and N (possible) local crops as a single tensor.

    Args:
        backbone: Backbone model.
        head: Head model.
    """

    def __init__(self, backbone: nn.Module, head: nn.Module):
        super().__init__()
        self.backbone = backbone
        self.head = head

    def forward(self, x):
        n_crops = len(x)
        # (n_samples * n_crops, 3, size, size)
        concatenated = torch.cat(x, dim=0)
        # (n_samples * n_crops, in_dim)
        embedding = self.backbone(concatenated)
        logits = self.head(embedding)  # (n_samples * n_crops, out_dim)
        chunks = logits.chunk(n_crops)  # n_crops * (n_samples, out_dim)

        return chunks



import shutil
from pathlib import Path

import cv2
import pytest

from quadra.utils.tests.helpers import _random_image


def _build_imagenette_dataset(tmp_path: Path, classes: int, class_samples: int) -> str:
    """Generate imagenette dataset in the format required by efficient_ad model.

    Args:
        tmp_path: Path to temporary directory
        classes: Number of mock imagenette classes
        class_samples: Number of samples for each mock imagenette class

    Returns:
        Path to imagenette dataset
    """
    parent_path = tmp_path / "imagenette2"
    parent_path.mkdir()
    train_path = parent_path / "train"
    train_path.mkdir()
    val_path = parent_path / "val"
    val_path.mkdir()

    for split in [train_path, val_path]:
        for i in range(classes):
            cl_path = split / f"class_{i}"
            cl_path.mkdir()
            for j in range(class_samples):
                image = _random_image()
                image_path = cl_path / f"fake_{j}.png"
                cv2.imwrite(str(image_path), image)

    return parent_path


@pytest.fixture
def imagenette_dataset(tmp_path: Path) -> str:
    """Generate a mock imagenette dataset to test efficient_ad model.

    Args:
        tmp_path: Path to temporary directory

    Yields:
        Path to imagenette dataset folder
    """
    yield _build_imagenette_dataset(tmp_path, classes=3, class_samples=3)

    if tmp_path.exists():
        shutil.rmtree(tmp_path)



from __future__ import annotations

import sklearn
import timm
import torch
import torch.nn.functional as F

from quadra.modules.base import SSLModule


class IDMM(SSLModule):
    """IDMM model.

    Args:
        model: backbone model
        prediction_mlp: student prediction MLP
        criterion: loss function
        multiview_loss: whether to use the multiview loss as definied in https://arxiv.org/abs/2201.10728.
            Defaults to True.
        mixup_fn: the mixup/cutmix function to be applied to a batch of images.
            Defaults to None.
        classifier: Standard sklearn classifier
        optimizer: optimizer of the training. If None a default Adam is used.
        lr_scheduler: lr scheduler. If None a default ReduceLROnPlateau is used.
        lr_scheduler_interval: interval at which the lr scheduler is updated.
    """

    def __init__(
        self,
        model: torch.nn.Module,
        prediction_mlp: torch.nn.Module,
        criterion: torch.nn.Module,
        multiview_loss: bool = True,
        mixup_fn: timm.data.Mixup | None = None,
        classifier: sklearn.base.ClassifierMixin | None = None,
        optimizer: torch.optim.Optimizer | None = None,
        lr_scheduler: object | None = None,
        lr_scheduler_interval: str | None = "epoch",
    ):
        super().__init__(
            model,
            criterion,
            classifier,
            optimizer,
            lr_scheduler,
            lr_scheduler_interval,
        )
        # self.save_hyperparameters()
        self.prediction_mlp = prediction_mlp
        self.mixup_fn = mixup_fn
        self.multiview_loss = multiview_loss

    def forward(self, x):
        z = self.model(x)
        p = self.prediction_mlp(z)
        return z, p

    def training_step(self, batch, batch_idx):
        # pylint: disable=unused-argument
        # Compute loss
        if self.multiview_loss:
            im_x, im_y, target = batch

            # Contrastive loss
            za, _ = self(im_x)
            zb, _ = self(im_y)
            za = F.normalize(za, dim=-1)
            zb = F.normalize(zb, dim=-1)
            s_aa = za.T @ za
            s_ab = za.T @ zb
            contrastive = (
                torch.log(torch.exp(s_aa).sum(-1))
                - torch.diagonal(s_aa)
                + torch.log(torch.exp(s_ab).sum(-1))
                - torch.diagonal(s_ab)
            )

            # Instance discrimination
            if self.mixup_fn is not None:
                im_x, target = self.mixup_fn(im_x, target)
            _, pred = self(im_x)
            loss = self.criterion(pred, target) + contrastive.mean()
        else:
            im_x, target = batch
            if self.mixup_fn is not None:
                im_x, target = self.mixup_fn(im_x, target)
            pred = self(im_x)
            loss = self.criterion(pred, target)

        self.log(
            "loss",
            loss,
            on_epoch=True,
            on_step=True,
            logger=True,
            prog_bar=True,
        )
        return loss



from __future__ import annotations

import json
import os
from typing import Any

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from label_studio_converter.brush import mask2rle
from omegaconf import DictConfig
from sklearn.metrics import ConfusionMatrixDisplay
from torch.utils.data import DataLoader

from quadra.utils import utils
from quadra.utils.patch.visualization import plot_patch_reconstruction
from quadra.utils.visualization import UnNormalize, plot_classification_results

log = utils.get_logger(__name__)


def save_classification_result(
    results: pd.DataFrame,
    output_folder: str,
    confusion_matrix: pd.DataFrame | None,
    accuracy: float,
    test_dataloader: DataLoader,
    reconstructions: list[dict],
    config: DictConfig,
    output: DictConfig,
    ignore_classes: list[int] | None = None,
):
    """Save classification results.

    Args:
        results: Dataframe containing the classification results
        output_folder: Folder where to save the results
        confusion_matrix: Confusion matrix
        accuracy: Accuracy of the model
        test_dataloader: Dataloader used for testing
        reconstructions: List of dictionaries containing polygons or masks
        config: Experiment configuration
        output: Output configuration
        ignore_classes: Eventual classes to ignore during reconstruction plot. Defaults to None.
    """
    # Save csv
    results.to_csv(os.path.join(output_folder, "test_results.csv"), index_label="index")

    if confusion_matrix is not None:
        # Save confusion matrix
        disp = ConfusionMatrixDisplay(
            confusion_matrix=np.array(confusion_matrix),
            display_labels=[x.replace("pred:", "") for x in confusion_matrix.columns.to_list()],
        )
        disp.plot(include_values=True, cmap=plt.cm.Greens, ax=None, colorbar=False, xticks_rotation=90)
        plt.title(f"Confusion Matrix (Accuracy: {(accuracy * 100):.2f}%)")
        plt.savefig(
            os.path.join(output_folder, "test_confusion_matrix.png"),
            bbox_inches="tight",
            pad_inches=0,
            dpi=300,
        )
        plt.close()

    if output.example:
        if not hasattr(test_dataloader.dataset, "idx_to_class"):
            raise ValueError("The provided dataset does not have an attribute 'idx_to_class")

        idx_to_class = test_dataloader.dataset.idx_to_class

        # Get misclassified samples
        example_folder = os.path.join(output_folder, "example")
        if not os.path.isdir(example_folder):
            os.makedirs(example_folder)

        # Skip if no no ground truth is available
        if not all(results["real_label"] == -1):
            for v in np.unique([results["real_label"], results["pred_label"]]):
                if v == -1:
                    continue

                k = idx_to_class[v]

                if ignore_classes is not None and v in ignore_classes:
                    continue

                plot_classification_results(
                    test_dataloader.dataset,
                    unorm=UnNormalize(mean=config.transforms.mean, std=config.transforms.std),
                    pred_labels=results["pred_label"].to_numpy(),
                    test_labels=results["real_label"].to_numpy(),
                    class_name=k,
                    original_folder=example_folder,
                    idx_to_class=idx_to_class,
                    pred_class_to_plot=v,
                    what="con",
                    rows=output.get("rows", 3),
                    cols=output.get("cols", 2),
                    figsize=output.get("figsize", (20, 20)),
                )

                plot_classification_results(
                    test_dataloader.dataset,
                    unorm=UnNormalize(mean=config.transforms.mean, std=config.transforms.std),
                    pred_labels=results["pred_label"].to_numpy(),
                    test_labels=results["real_label"].to_numpy(),
                    class_name=k,
                    original_folder=example_folder,
                    idx_to_class=idx_to_class,
                    pred_class_to_plot=v,
                    what="dis",
                    rows=output.get("rows", 3),
                    cols=output.get("cols", 2),
                    figsize=output.get("figsize", (20, 20)),
                )

        for counter, reconstruction in enumerate(reconstructions):
            is_polygon = True
            if isinstance(reconstruction["prediction"], np.ndarray):
                is_polygon = False

            if is_polygon:
                if len(reconstruction["prediction"]) == 0:
                    continue
            elif reconstruction["prediction"].sum() == 0:
                continue

            if counter > 5:
                break

            to_plot = plot_patch_reconstruction(
                reconstruction,
                idx_to_class,
                class_to_idx=test_dataloader.dataset.class_to_idx,  # type: ignore[attr-defined]
                ignore_classes=ignore_classes,
                is_polygon=is_polygon,
            )

            if to_plot:
                output_name = f"reconstruction_{os.path.splitext(os.path.basename(reconstruction['file_name']))[0]}.png"
                plt.savefig(os.path.join(example_folder, output_name), bbox_inches="tight", pad_inches=0)

            plt.close()


class RleEncoder(json.JSONEncoder):
    """Custom encoder to convert numpy arrays to RLE."""

    def default(self, o: Any):
        """Customize standard encoder behaviour to convert numpy arrays to RLE."""
        if isinstance(o, np.ndarray):
            return mask2rle(o)
        return json.JSONEncoder.default(self, o)



from .backbones import BaseNetworkBuilder, TimmNetworkBuilder, TorchHubNetworkBuilder, TorchVisionNetworkBuilder

__all__ = [
    "BaseNetworkBuilder",
    "TorchVisionNetworkBuilder",
    "TorchHubNetworkBuilder",
    "TimmNetworkBuilder",
]
